# 'init' subcommand: get user's preferences and write config file

import sys
import os
import importlib.metadata
import re
import logging
from . import ncbi_helper
from . import nextclade_helper
from . import config


def get_input(prompt):
    """Get input from the user.  Quit gracefully if it looks like the user wants out."""
    try:
        text = input(prompt)
    except (KeyboardInterrupt, EOFError):
        print("")
        sys.exit(1)
    if re.match(r"^[Qq](uit)?", text):
        print("Exiting...")
        sys.exit(0)
    return text


def prompt_int_choice(prompt, min_value=1, max_value=None):
    """Get an integer choice from the user, with optional min and max values.
    min_value will be presented as default in case of empty input."""
    while True:
        try:
            choice = get_input(f"{prompt} [{min_value}]: ")
            choice = int(choice) if choice else min_value
            if min_value <= choice and (max_value is None or choice <= max_value):
                return choice
            else:
                print(f"Please enter a number between {min_value} and {max_value}.")
        except ValueError:
            print("Invalid input. Please enter a number.")


def prompt_taxonomy_id(species_term, tax_entries):
    """Given a list of NCBI Datasets taxonomy match records, display to the user and ask them to choose one,
    also giving an option for None of the above / go back.  Return 1-based index of user's choice."""
    print(f"\nFound the following matches for '{species_term}':")
    for idx, entry in enumerate(tax_entries):
        print(f"{idx+1}. {entry['sci_name']}")
    print(f"{len(tax_entries)+1}. None of the above, go back")
    return prompt_int_choice("Enter a number to choose an option", 1, len(tax_entries)+1)


def prompt_refseq_id(taxid, refseq_entries):
    """Given a list of NCBI RefSeq records, display to the user and ask them to choose one,
    also giving an option for None of the above / go back.  Return 1-based index of user's choice."""
    print(f"\nFound the following RefSeq IDs for Taxonomy ID {taxid}:")
    for idx, entry in enumerate(refseq_entries):
        label = f"{entry['accession']}: {entry['title']}"
        if (entry["strain"] and entry["strain"] != "No strain" and entry["strain"] not in label):
            label += f" ({entry['strain']})"
        print(f"{idx+1}. {label}")
    print(f"{len(refseq_entries)+1}. None of the above, go back")
    return prompt_int_choice("Enter a number to choose an option", 1, len(refseq_entries)+1)


def get_species_taxonomy_refseq(ncbi):
    """Prompt user for a species search term.  Look up Taxonomy IDs for species,
    prompt the user to select a Taxonomy ID, look up RefSeqs for the selected Taxonomy ID,
    prompt the user to select a RefSeq.  Return Taxonomy ID, RefSeq accession and RefSeq Assembly ID."""
    species_term = ""
    while not species_term:
        species_term = get_input("\nWhat is your virus of interest? ")
    return lookup_taxonomy_refseq(ncbi, species_term)


def lookup_taxonomy_refseq(ncbi, species_term):
    """Look up Taxonomy IDs for species, prompt the user to select a Taxonomy ID,
    look up RefSeqs for the selected Taxonomy ID, prompt the user to select a RefSeq.
    Return Taxonomy ID, RefSeq accession and RefSeq Assembly ID."""
    print(f"Looking up NCBI Taxonomy entries for '{species_term}'...")
    tax_entries = ncbi.get_taxonomy_entries('"' + species_term + '"')
    if tax_entries:
        return get_taxonomy_refseq(ncbi, species_term, tax_entries)
    else:
        print(f"\nNo matches found for '{species_term}'.")
        return get_species_taxonomy_refseq(ncbi)


def get_taxonomy_refseq(ncbi, species_term, tax_entries):
    """Prompt the user to select a Taxonomy ID, look up RefSeqs for the selected Taxonomy ID,
    prompt the user to select a RefSeq.  Return Taxonomy ID, RefSeq accession and RefSeq Assembly ID."""
    choice = prompt_taxonomy_id(species_term, tax_entries)
    if choice > len(tax_entries):
        # User says none of the above, go back.
        return get_species_taxonomy_refseq(ncbi)
    else:
        taxid = tax_entries[choice-1]["tax_id"]
        return lookup_refseq(ncbi, species_term, tax_entries, taxid)


def lookup_refseq(ncbi, species_term, tax_entries, taxid):
    """Look up RefSeqs for the selected Taxonomy ID, prompt the user to select a RefSeq.
    Return Taxonomy ID and RefSeq ID."""
    print(f"Looking up RefSeqs associated with Taxonomy ID {taxid}...")
    refseq_entries = ncbi.get_refseqs_for_taxid(taxid)
    if refseq_entries:
        return get_refseq(ncbi, species_term, tax_entries, taxid, refseq_entries)
    else:
        print(f"No RefSeqs found for Taxonomy ID '{taxid}'.  Try a different match from species search.")
        return get_taxonomy_refseq(ncbi, species_term, tax_entries)


def get_refseq(ncbi, species_term, tax_entries, taxid, refseq_entries):
    """Prompt the user to select a RefSeq.  Return Taxonomy ID, RefSeq accession and RefSeq Assembly ID."""
    choice = prompt_refseq_id(taxid, refseq_entries)
    if choice > len(refseq_entries):
        # User says none of the above, go back.
        return get_taxonomy_refseq(ncbi, species_term, tax_entries)
    else:
        refseq_id = refseq_entries[choice-1]["accession"]
        print(f"Looking up NCBI Assembly accession for selected RefSeq '{refseq_id}'...")
        assembly_id = ncbi.get_assembly_acc_for_refseq_acc(refseq_id)
        if not assembly_id:
            print(f"Could not find assembly ID for RefSeq ID {refseq_id} -- can't download RefSeq.")
            print("Try choosing a different RefSeq.")
            return get_refseq(ncbi, species_term, tax_entries, taxid, refseq_entries)
        return species_term, taxid, refseq_id, assembly_id


def clade_columns_from_dataset(nextclade_dataset):
    """Return a comma-sep string of clade column names"""
    return ','.join(nextclade_dataset.get("clades", {}).keys())


def prompt_nextclade_path_columns(nextclade_datasets, species, is_interactive):
    """Prompt the user to select a Nextclade dataset from a list of paths."""
    if is_interactive:
        print(f"\nFound the following Nextclade datasets for '{species}':")
        for idx, dataset in enumerate(nextclade_datasets):
            print(f"{idx+1}. {dataset['path']}: {dataset['name']}")
        print(f"{len(nextclade_datasets)+1}. None of the above")
        choice = prompt_int_choice("Enter a number to choose an option", 1, len(nextclade_datasets)+1)
        if choice > len(nextclade_datasets):
            return "", ""
        else:
            dataset = nextclade_datasets[choice-1]
            return dataset["path"], clade_columns_from_dataset(dataset)
    else:
        print(f"Multiple Nextclade datasets found for '{species}':")
        for dataset in nextclade_datasets:
            print(f" - {dataset['path']}: {dataset['name']}")
        print("Please specify one of these using the -x/--nextclade_dataset option.")
        sys.exit(1)


def search_nextclade_datasets(datasets, term):
    """Search for Nextclade datasets that match the given term."""
    matches = []
    term_lower = term.lower()
    for dataset in datasets:
        if term_lower in dataset["name"].lower() or term_lower in dataset["path"].lower():
            matches.append(dataset)
    return matches


def get_nextclade_path_columns(args_nextclade, species, is_interactive):
    """Return a list of Nextclade dataset paths whose names match arg or species."""
    print(f"Looking up Nextclade datasets for '{species}'...")
    datasets = nextclade_helper.nextclade_get_index()
    if args_nextclade == "":
        return "", ""
    elif args_nextclade:
        for dataset in datasets:
            if dataset["path"] == args_nextclade:
                print(f"Using Nextclade dataset '{dataset['path']}'")
                return dataset['path'], clade_columns_from_dataset(dataset)
        print(f"Nextclade dataset '{args_nextclade}' not found. Available datasets:")
        for dataset in datasets:
            print(f"- {dataset['path']}: {dataset['name']}")
        print("Please try again with a different value for -x/--nextclade_dataset.")
        sys.exit(1)
    elif species:
        matches = search_nextclade_datasets(datasets, species)
        if len(matches) == 0 and ' ' in species:
            # Search each word in species separately; quit if a word has matches
            for word in species.lower().split(' '):
                if word in ["human", "virus", "fever", "genotype"] or len(word) < 3:
                    continue
                word_matches = search_nextclade_datasets(datasets, word)
                if word_matches:
                    matches = word_matches
                    break
        if len(matches) == 0:
            print(f"No Nextclade datasets found for '{species}'.")
            return "", ""
        elif len(matches) == 1:
            print(f"Found Nextclade dataset for '{species}': {matches[0]['path']}")
            return matches[0]["path"], clade_columns_from_dataset(matches[0])
        else:
            matches.reverse()
            return prompt_nextclade_path_columns(matches, species, is_interactive)
    else:
        return "", ""


def check_proportion(proportion):
    """Make sure the given value, while a string value, converts to a float between 0 and 1.  Empty/None is okay."""
    if proportion:
        try:
            value = float(proportion)
        except ValueError:
            return False, f"Sorry, need a value between 0 and 1, '{proportion}' doesn't parse"
        if value < 0:
            return False, "Sorry, need a value between 0 and 1, not a negative number"
        if value > 1.0:
            return False, f"Sorry, need a value between 0 and 1, {proportion} is too large"
    return True, None


def check_int(val, min=None, max=None):
    """Make sure the given value converts to an integer.  Empty/None is okay."""
    if val:
        try:
            value = int(val)
        except ValueError:
            return False, f"Sorry, need an integer, '{val}' doesn't parse"
        if min is not None:
            if value < min:
                return False, f"Sorry, need at least {min}, {val} is too low."
        if max is not None:
            if value > max:
                return False, f"Sorry, the maximum is {max}, {val} is too large."
    return True, None


def check_nonnegative_int(val):
    return check_int(val, min=0)


def check_optional_file_readable(filename):
    """If filename is empty, that's fine; if non-empty, make sure it's a readable file"""
    if filename:
        if not os.path.exists(filename):
            return False, f"Sorry, file '{filename}' does not exist"
        if not os.path.isfile(filename):
            return False, f"Sorry, file '{filename}' does not appear to be a regular file"
        else:
            try:
                f = open(filename, "r")
                f.close()
            except IOError:
                return False, f"Sorry, unable to read file '{filename}'."
    return True, None


def check_dir_exists_or_creatable(dirname):
    """If dirname exists, all good; otherwise try to create it.  Return False and error message if unable."""
    if not os.path.exists(dirname):
        print(f"Creating directory {dirname}...")
        try:
            os.makedirs(dirname)
        except OSError:
            return False, f"Sorry, can't create directory path '{dirname}', please enter a different one."
    elif not os.path.isdir(dirname):
        return False, f"Sorry, '{dirname}' exists but is not a directory, please enter a different name."
    return True, None


def prompt_with_checker(prompt, default, check_function):
    """Prompt the user for a value and check their answer.  If there's a problem then explain
    and prompt again until their answer meets criteria."""
    ok = False
    answer = None
    while not ok:
        if default is not None:
            answer = get_input(f"\n{prompt} [{default}]: ")
            if not answer:
                answer = default
        else:
            answer = get_input(f"\n{prompt}: ")
        ok, error_message = check_function(answer)
        if not ok:
            print(error_message)
    return answer


def check_write_config(config_contents, config_path):
    """Return True if able to write config to config_path, False otherwise."""
    try:
        config.write_config(config_contents, config_path)
        return True
    except (NotADirectoryError, OSError):
        return False


def get_min_length_proportion(args_min_length_proportion, is_interactive):
    min_length_proportion = config.DEFAULT_MIN_LENGTH_PROPORTION
    if args_min_length_proportion:
        min_length_proportion = args_min_length_proportion
        ok, error_message = check_proportion(min_length_proportion)
        if not ok:
            print(f"{error_message}\nPlease try again with a different value for --min_length_proportion", file=sys.stderr)
            sys.exit(1)
    elif is_interactive:
        min_length_proportion = prompt_with_checker("GenBank sequences will be filtered to retain only those whose length is at least this proportion of the RefSeq length",
                                                    min_length_proportion, check_proportion)
    return str(min_length_proportion)


def get_max_N_proportion(args_max_N_proportion, is_interactive):
    max_N_proportion = config.DEFAULT_MAX_N_PROPORTION
    if args_max_N_proportion:
        max_N_proportion = args_max_N_proportion
        ok, error_message = check_proportion(max_N_proportion)
        if not ok:
            print(f"{error_message}\nPlease try again with a different value for --max_N_proportion", file=sys.stderr)
            sys.exit(1)
    elif is_interactive:
        max_N_proportion = prompt_with_checker("GenBank sequences will be filtered to retain only those with at most this proportion of 'N' or ambiguous bases",
                                               max_N_proportion, check_proportion)
    return str(max_N_proportion)


def get_max_parsimony(args_max_parsimony, is_interactive):
    max_parsimony = config.DEFAULT_MAX_PARSIMONY
    if args_max_parsimony:
        max_parsimony = args_max_parsimony
        ok, error_message = check_nonnegative_int(max_parsimony)
        if not ok:
            print(f"{error_message}\nPlease try again with a different value for --max_parsimony", file=sys.stderr)
            sys.exit(1)
    elif is_interactive:
        max_parsimony = prompt_with_checker("After the tree is constructed, sequences with more than this number of private substitutions will be removed",
                                            max_parsimony, check_nonnegative_int)
    return str(max_parsimony)


def get_max_branch_length(args_max_branch_length, is_interactive):
    max_branch_length = config.DEFAULT_MAX_BRANCH_LENGTH
    if args_max_branch_length:
        max_branch_length = args_max_branch_length
        ok, error_message = check_nonnegative_int(max_branch_length)
        if not ok:
            print(f"{error_message}\nPlease try again with a different value for --max_branch_length", file=sys.stderr)
            sys.exit(1)
    elif is_interactive:
        max_branch_length = prompt_with_checker("After the tree is constructed, branches with more than this number of substitutions defining the branch will be removed",
                                                max_branch_length, check_nonnegative_int)
    return str(max_branch_length)


def get_user_fasta(args_fasta, is_interactive):
    fasta = ''
    if args_fasta is not None:
        fasta = args_fasta
        ok, error_message = check_optional_file_readable(fasta)
        if not ok:
            print(f"{error_message}\nPlease try again with a different file for --fasta.", file=sys.stderr)
            sys.exit(1)
    elif is_interactive:
        fasta = prompt_with_checker("If you have your own fasta file, then enter its path", "", check_optional_file_readable)
    return fasta


def get_workdir(args_workdir, is_interactive):
    if args_workdir:
        workdir = args_workdir
        ok, error_message = check_dir_exists_or_creatable(workdir)
        if not ok:
            print(f"{error_message}\nPlease try again with a different path for --workdir.", file=sys.stderr)
            sys.exit(1)
    else:
        workdir = prompt_with_checker("Enter directory where sequences should be downloaded and trees built", ".", check_dir_exists_or_creatable)
    return workdir


def make_config(config_contents, workdir, refseq_id, taxid, args_config, is_interactive):
    config_path_default = f"{workdir}/viral_usher_config_{refseq_id}_{taxid}.toml"
    if args_config:
        config_path = args_config
        if not check_write_config(config_contents, config_path):
            print(f"Unable to write to --config {config_path}.  Please try again with a different --config path.")
            sys.exit(1)
    elif is_interactive:
        config_ok = False
        while not config_ok:
            config_path = get_input(f"\nEnter path for config file [{config_path_default}]: ") or config_path_default
            config_ok = check_write_config(config_contents, config_path)
            if not config_ok:
                print(f"Unable to write config to {config_path}.")
            else:
                print(f"Wrote config file to {config_path}")
    else:
        config_path = config_path_default
        if not check_write_config(config_contents, config_path):
            print(f"Unable to write to default config path {config_path}.  Please try again with a different --config path.")
            sys.exit(1)
    return config_path


def handle_init(args):
    logging.basicConfig(level=logging.INFO, format='%(asctime)s - %(levelname)s - %(message)s')
    ncbi = ncbi_helper.NcbiHelper()
    # First sort out the interdependent refseq and taxonomy id options.
    if args.refseq:
        refseq_id = args.refseq
        assembly_id = ncbi.get_assembly_acc_for_refseq_acc(refseq_id)
        if not assembly_id:
            print(f"Could not find assembly ID for RefSeq ID {refseq_id} -- can't download RefSeq.")
            sys.exit(1)
        species = args.species
        taxid = args.taxonomy_id
        if not species or not taxid:
            species, taxid = ncbi.get_species_taxid_for_refseq(refseq_id)
        if not species and args.species:
            species = args.species
        if not taxid and args.taxonomy_id:
            taxid = args.taxonomy_id
        if not species or not taxid:
            print(f"Could not find species or taxonomy ID for RefSeq ID {refseq_id}.", file=sys.stderr)
            sys.exit(1)
        taxid = ncbi.get_species_level_taxid(taxid)
        taxid = str(taxid)
        if args.taxonomy_id:
            if taxid != args.taxonomy_id:
                if taxid is None:
                    print(f"No taxid for refseq GI for {refseq_id}")
                # TODO: it's not a problem if the two taxids have an ancestor-descendant relationship -- look it up.
                print(f"\nNOTE: RefSeq ID {refseq_id} is associated with species-level Taxonomy ID {taxid}, not the provided Taxonomy ID {args.taxonomy_id}.\n")
                taxid = args.taxonomy_id
    else:
        if args.taxonomy_id:
            species, taxid, refseq_id, assembly_id = lookup_refseq(ncbi, args.species, [], args.taxonomy_id)
        elif args.species:
            species, taxid, refseq_id, assembly_id = lookup_taxonomy_refseq(ncbi, args.species)
        else:
            species, taxid, refseq_id, assembly_id = get_species_taxonomy_refseq(ncbi)
        taxid = ncbi.get_species_level_taxid(taxid)

    # If --refseq and --workdir are given then accept defaults for other options
    is_interactive = not (args.refseq and args.workdir)

    nextclade_path, nextclade_columns = get_nextclade_path_columns(args.nextclade_dataset, species, is_interactive)
    min_length_proportion = get_min_length_proportion(args.min_length_proportion, is_interactive)
    max_N_proportion = get_max_N_proportion(args.max_N_proportion, is_interactive)
    max_parsimony = get_max_parsimony(args.max_parsimony, is_interactive)
    max_branch_length = get_max_branch_length(args.max_branch_length, is_interactive)
    fasta = get_user_fasta(args.fasta, is_interactive)
    workdir = get_workdir(args.workdir, is_interactive)

    viral_usher_version = importlib.metadata.version('viral_usher')
    config_contents = {
        "viral_usher_version": viral_usher_version,
        "refseq_acc": refseq_id,
        "refseq_assembly": assembly_id,
        "species": species,
        "taxonomy_id": taxid,
        "nextclade_dataset": nextclade_path,
        "nextclade_clade_columns": nextclade_columns,
        "min_length_proportion": min_length_proportion,
        "max_N_proportion": max_N_proportion,
        "max_parsimony": max_parsimony,
        "max_branch_length": max_branch_length,
        "extra_fasta": fasta,
        "workdir": os.path.abspath(workdir),
    }
    config_path = make_config(config_contents, workdir, refseq_id, taxid, args.config, is_interactive)

    print(f"\nReady to roll!  Next, try running 'viral_usher build --config {config_path}'\n")



import argparse
import importlib.metadata
import sys
from .init import handle_init
from .build import handle_build
from . import config


def main():
    viral_usher_version = importlib.metadata.version('viral_usher')
    parser = argparse.ArgumentParser(
        prog="viral_usher",
        description=f"viral_usher {viral_usher_version} — Build viral phylogenies from NCBI data.",
        formatter_class=argparse.RawDescriptionHelpFormatter
        )
    parser.add_argument("-v", "--version", action="store_true", help="display viral_usher version and exit")
    subparsers = parser.add_subparsers(dest="command")

    # Init subcommand
    init_parser = subparsers.add_parser("init", help="Initialize configuration for a new species/taxon and reference")
    init_parser.add_argument("-r", "--refseq", help="RefSeq accession to use as reference/root, if known")
    init_parser.add_argument("-t", "--taxonomy_id", help="NCBI Taxonomy ID of the viral species, if known")
    init_parser.add_argument("-s", "--species", help="Viral species name (if Taxonomy ID is not known)")
    init_parser.add_argument("-f", "--fasta", help="Additional sequences to include in tree")
    init_parser.add_argument("-x", "--nextclade_dataset", help="Nextclade dataset path (e.g. 'nextstrain/dengue/all') (default: auto-detect)")
    init_parser.add_argument("-l", "--min_length_proportion", type=float, help=f"Minimum proportion of RefSeq length to require for GenBank sequences (default: {config.DEFAULT_MIN_LENGTH_PROPORTION})")
    init_parser.add_argument("-n", "--max_N_proportion", type=float, help=f"Maximum proportion of N bases to allow in GenBank sequences (default: {config.DEFAULT_MAX_N_PROPORTION})")
    init_parser.add_argument("-a", "--max_parsimony", type=int, help=f"Remove sequences from the tree with parsimony score (private substitution count) greater than this (default: {config.DEFAULT_MAX_PARSIMONY})")
    init_parser.add_argument("-b", "--max_branch_length", type=int, help=f"Remove branches from the tree with branch length (substitution count) greater than this (default: {config.DEFAULT_MAX_BRANCH_LENGTH})")
    init_parser.add_argument("-w", "--workdir", help="Directory in which tree files will be built")
    init_parser.add_argument("-c", "--config", type=str, help="Path to config file output")

    # Build subcommand
    build_parser = subparsers.add_parser("build", help="Run the pipeline to download sequences and build a tree")
    build_parser.add_argument("-c", "--config", type=str, required=True, help="Path to config file input")
    build_parser.add_argument("-d", "--docker_image", type=str, help=f"Use this docker image instead of docker.io/{config.DEFAULT_DOCKER_IMAGE}")
    build_parser.add_argument("-u", "--update", action="store_true", help="Update an existing tree with new sequences (requires optimized.pb.gz tree from previous build)")

    args = parser.parse_args()

    # Parse known args to check for -h or -v
    args, unknown = parser.parse_known_args()

    # Check for -v/--version
    if args.version:
        print(f"viral_usher {viral_usher_version}")
        sys.exit(0)

    # If -h/--help is present, let argparse print help and exit
    if "-h" in sys.argv or "--help" in sys.argv:
        print(f"unknown is {unknown}\n\n", file=sys.stderr)
        parser.parse_args()

    # A subcommand is required unless -h or -v were given
    if not args.command:
        parser.error("A subcommand {init,build} is required unless -h/--help or -v/--version is given.")

    # Parse all args for subcommand handlers
    args = parser.parse_args()

    if args.command == "init":
        handle_init(args)
    elif args.command == "build":
        handle_build(args)
    else:
        parser.print_help()
        exit(1)


if __name__ == "__main__":
    main()



# 'build' subcommand: build a tree according to parameters in config file

import datetime
import docker
import filecmp
import importlib.metadata
import os
import shutil
import subprocess
import sys
from . import config

docker_platform = 'linux/amd64'
docker_workdir = '/data'


def check_docker_command():
    """Make sure we can run the docker command and connect to the docker server.  If there's a problem, return False with
    an error message; otherwise return True with no message."""
    try:
        subprocess.run(['docker', '--help'], check=True, capture_output=True)
    except (subprocess.CalledProcessError, FileNotFoundError):
        return False, "Unable to run 'docker --help' -- please install Docker from https://www.docker.com/ and try again."
    try:
        subprocess.run(['docker', 'images'], check=True, capture_output=True)
    except subprocess.CalledProcessError:
        return False, "Unable to run 'docker images' -- please make sure the docker daemon is running (e.g. on a Mac, start the Docker app)"
    return True, None


def pull_docker_image(client: docker.DockerClient, image_name: str):
    """Pull the docker image corresponding to this version of the viral_usher python package unless it is already present."""
    try:
        client.images.pull(image_name, platform=docker_platform)
    except (docker.errors.ContainerError) as e:
        print(f"Failed to pull docker container {image_name}:\n{e}", file=sys.stderr)
        sys.exit(1)
    except (TimeoutError):
        print("Timeout error while trying to pull docker image; maybe try again later?")
        sys.exit(1)


def get_docker_image(client: docker.DockerClient, args_docker_image: str):
    if args_docker_image:
        try:
            client.images.get(args_docker_image)
            print(f"Using docker image {args_docker_image}")
        except docker.errors.ImageNotFound:
            print(f"Docker image '{args_docker_image}' not found.  Please try again with a different --docker_image value.", sys.stderr)
        return args_docker_image
    else:
        viral_usher_version = importlib.metadata.version('viral_usher')
        image_name = config.DEFAULT_DOCKER_IMAGE + ':v' + viral_usher_version
        try:
            client.images.get(image_name)
            print(f"Using docker image {image_name}")
        except docker.errors.ImageNotFound:
            print(f"Docker image {image_name} is not present, pulling it from DockerHub...")
            pull_docker_image(client, image_name)
        return image_name


def maybe_copy_to_workdir(filepath, workdir):
    """Return filepath's relative path within workdir, copying the file into workdir if necessary."""
    basename = os.path.basename(filepath)
    dirname = os.path.dirname(filepath)
    dirname_abs = os.path.abspath(dirname)
    workdir_abs = os.path.abspath(workdir)
    if dirname_abs == workdir_abs:
        return basename
    else:
        if dirname_abs.startswith(workdir_abs + '/'):
            relpath = dirname_abs.removeprefix(workdir_abs + '/')
            relname = relpath + '/' + basename
            return relname
        else:
            workdir_file = workdir + '/' + basename
            if os.path.exists(workdir_file):
                if not filecmp.cmp(filepath, workdir_file, shallow=False):
                    print(f"File {filepath} is specified, but workdir file {workdir_file} already exists and is not the same", file=sys.stderr)
                    sys.exit(1)
            else:
                shutil.copy(filepath, workdir)
            return basename


def rewrite_config(config_contents, workdir):
    """Make a new timestamped config file in workdir that contains the given config settings"""
    now = datetime.datetime.now()
    new_name = 'local_config.' + now.strftime("%Y-%m-%d_%H-%M-%S") + '.toml'
    new_path = workdir + '/' + new_name
    config.write_config(config_contents, new_path)
    return new_name


def localize_config(args_config, config_contents):
    """When running in docker, only the workdir will be available (as the current directory).
    So absolute paths in the config need to be converted to relative paths, and if input files
    are not already in the workdir, then they need to be copied there."""
    workdir = config_contents['workdir']
    # If extra_fasta is given, make sure it will be available in workdir and that
    # the config used by the docker image script uses its relative path in workdir.
    config_changed = False
    extra_fasta = config_contents.get('extra_fasta', '')
    if extra_fasta:
        extra_fasta_rel = maybe_copy_to_workdir(extra_fasta, workdir)
        if extra_fasta_rel != extra_fasta:
            config_contents['extra_fasta'] = extra_fasta_rel
            config_changed = True
    if config_changed:
        config_rel = rewrite_config(config_contents, workdir)
    else:
        config_rel = maybe_copy_to_workdir(args_config, workdir)
    return config_rel


def run_in_docker(workdir, config_rel, args_docker_image, args_update):
    """Start up a docker container that runs the pipeline."""
    docker_client = docker.from_env()
    user_id = os.getuid()
    group_id = os.getgid()
    docker_image = get_docker_image(docker_client, args_docker_image)
    command = ["viral_usher_build", "--config", config_rel]
    if args_update:
        command.append("--update")
    try:
        container = docker_client.containers.run(
            docker_image,
            platform=docker_platform,
            remove=True,
            network_mode='host',
            user=':'.join([str(user_id), str(group_id)]),
            volumes=[':'.join([workdir, docker_workdir])],
            command=command,
            detach=True,  # Run container in background so we can stream logs
            stdout=True,
            stderr=True,
            tty=True
        )
        # Stream logs in real time
        for line in container.logs(stream=True):
            print(line.decode(), end='')
        exit_code = container.wait()['StatusCode']
        if exit_code != 0:
            print(f"docker container {docker_image} failed with exit code {exit_code}", file=sys.stderr)
            sys.exit(exit_code)
    except (docker.errors.ContainerError) as e:
        print(f"docker container {docker_image} failed:\n{e}", file=sys.stderr)
        sys.exit(1)
    except (TimeoutError):
        print("Timeout error while trying to run docker; maybe try again later?")
        sys.exit(1)


def handle_build(args):
    """Set up input files in workdir, run pipeline in docker and check final results."""
    ok, error = check_docker_command()
    if not ok:
        print(f"\n{error}\n", file=sys.stderr)
        sys.exit(1)
    print(f"Running pipeline with config file: {args.config}")
    config_contents = config.parse_config(args.config)
    config_rel = localize_config(args.config, config_contents)
    workdir = config_contents['workdir']
    run_in_docker(workdir, config_rel, args.docker_image, args.update)
    if os.path.exists(f"{workdir}/tree.jsonl.gz"):
        print(f"Success -- you can view {workdir}/tree.jsonl.gz using https://taxonium.org/ now.")
    else:
        print(f"The pipeline exited without an error status but the expected result file {workdir}/tree.jsonl.gz is not there.  Please file an issue in GitHub: https://github.com/AngieHinrichs/viral_usher/issues/new", file=sys.stderr)
        sys.exit(1)



import pytest
import tempfile
from pathlib import Path
import viral_usher.config


@pytest.fixture(scope="session")
def written_config_file(config_zika_no_workdir_path, config_zika_no_workdir_data):
    """Create a config file once for the entire test session."""
    # Create temporary directory that persists for the session
    with tempfile.TemporaryDirectory() as temp_dir:
        # Temporary path for written config file
        config_path = Path(temp_dir) / "test_config.toml"

        # Test data to write: add missing workir (in temp_dir) to zika example
        config_data = config_zika_no_workdir_data
        config_data["workdir"] = Path(temp_dir) / "zika"

        # Write the config file
        viral_usher.config.write_config(config_data, config_path)

        # Verify it was written correctly
        assert config_path.exists()

        yield config_path
        # File gets cleaned up automatically when temp_dir exits


def test_check_refseq_acc_good():
    assert viral_usher.config.check_refseq_acc("NC_012345.1") is None
    assert viral_usher.config.check_refseq_acc("NZ_012345.1") is None
    assert viral_usher.config.check_refseq_acc("AC_012345.1") is None


def test_check_refseq_acc_bad():
    with pytest.raises(ValueError):
        viral_usher.config.check_refseq_acc(None)
    with pytest.raises(ValueError):
        viral_usher.config.check_refseq_acc("")
    with pytest.raises(ValueError):
        viral_usher.config.check_refseq_acc("AB034985")


def test_check_taxonomy_id_good():
    assert viral_usher.config.check_taxonomy_id("12345") is None


def test_check_taxonomy_id_bad():
    with pytest.raises(ValueError):
        viral_usher.config.check_taxonomy_id(None)
    with pytest.raises(ValueError):
        viral_usher.config.check_taxonomy_id("")
    with pytest.raises(ValueError):
        viral_usher.config.check_taxonomy_id("measles")


def test_check_refseq_assembly_good():
    assert viral_usher.config.check_refseq_assembly("GCA_123456789.1") is None
    assert viral_usher.config.check_refseq_assembly("GCF_123456789.1") is None


def test_check_refseq_assembly_bad():
    with pytest.raises(ValueError):
        viral_usher.config.check_refseq_assembly(None)
    with pytest.raises(ValueError):
        viral_usher.config.check_refseq_assembly("")
    with pytest.raises(ValueError):
        viral_usher.config.check_refseq_assembly("NC_012345.1")


def test_read_config_no_workdir(config_zika_no_workdir_path, config_zika_no_workdir_data):
    result = viral_usher.config.read_config(config_zika_no_workdir_path)
    assert result == config_zika_no_workdir_data


def test_read_config_file_not_found():
    with pytest.raises(FileNotFoundError):
        viral_usher.config.read_config("nonexistent.toml")


def test_write_config_creates_file(written_config_file):
    """Test that the config file was created and exists."""
    assert written_config_file.exists()
    assert written_config_file.suffix == ".toml"


def test_read_written_config(written_config_file):
    """Test reading the config file that was written."""
    config_data = viral_usher.config.read_config(written_config_file)
    assert config_data["refseq_acc"] == "NC_035889.1"
    assert config_data["taxonomy_id"] == "64320"



import argparse
import csv
import datetime
import gzip
import io
import json
import lzma
import os
import re
import shutil
import subprocess
import sys
import time
import zipfile
from Bio import SeqIO
from . import config
from . import ncbi_helper

update_tree_input = "optimized.pb.gz"
update_nextclade_input = "nextclade.clade.tsv"


def run_command(command, stdout_filename=None, stderr_filename=None, fail_ok=False):
    """Run a command, complain and raise if it fails (unless fail_ok, then just return False)"""
    success = False
    stdout = None
    stderr = None
    if stdout_filename:
        stdout = open(stdout_filename, 'w')
    if stderr_filename:
        stderr = open(stderr_filename, 'w')
    try:
        subprocess.run(command, check=True, stdout=stdout, stderr=stderr)
        success = True
    except subprocess.CalledProcessError as e:
        if fail_ok:
            pass
        else:
            if e.stderr:
                print(f"{command[0]} failed: {e.stderr}\nFailed command: {' '.join(command)}", file=sys.stderr)
            elif stderr_filename:
                print(f"{command[0]} failed, see {stderr_filename}\nFailed command: {' '.join(command)}", file=sys.stderr)
            else:
                print(f"{command[0]} failed\nFailed command: {' '.join(command)}", file=sys.stderr)
            sys.exit(1)
    if stdout:
        stdout.close()
    if stderr:
        stderr.close()
    return success


def run_command_get_stdout(command):
    """Run a command and return its stdout output."""
    try:
        result = subprocess.run(command, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        return result.stdout.decode('utf-8')
    except subprocess.CalledProcessError as e:
        print(f"Command failed: {e.stderr}\nFailed command: {' '.join(command)}", file=sys.stderr)
        sys.exit(1)


def start_timing(message):
    print(message)
    return time.perf_counter()


def finish_timing(start_time):
    elapsed_time = time.perf_counter() - start_time
    print(f"... done in {elapsed_time:.1f}s")


def normalize_segment(segment):
    # Many viruses have segments S (small), M (medium), and L (large).  The vast majority of records use S, M or L
    # but there are a few oddballs that spell it out or use middle instead of medium.  Change those to S/M/L.
    if segment and len(segment) > 1:
        seg_lower = segment.lower()
        if seg_lower == 'small':
            segment = 'S'
        elif seg_lower == 'large':
            segment = 'L'
        elif seg_lower == 'medium' or seg_lower == 'middle':
            segment = 'M'
    return segment


def unpack_refseq_zip(refseq_zip, refseq_acc):
    """Extract and rename the fasta and gbff files from the RefSeq zip."""
    start_time = start_timing(f"Unpacking {refseq_zip}...")
    with zipfile.ZipFile(refseq_zip, 'r') as zip_ref:
        fasta_path = 'refseq.fasta'
        gbff_path = 'refseq.gbff'
        length = 0
        fasta_found = False
        gbff_found = False
        for name in zip_ref.namelist():
            if name.endswith('.fna'):
                with io.TextIOWrapper(zip_ref.open(name, 'r'), encoding='utf-8') as fasta_in, open(fasta_path, 'w') as fasta_out:
                    for record in SeqIO.parse(fasta_in, 'fasta'):
                        if record.id == refseq_acc:
                            record.description = record.id
                            SeqIO.write(record, fasta_out, 'fasta')
                            length = len(record.seq)
                            fasta_found = True
            elif name.endswith('.gbff'):
                with zip_ref.open(name, 'r') as gbff_in, open(gbff_path, 'wb') as gbff_out:
                    shutil.copyfileobj(gbff_in, gbff_out)
                gbff_found = True
        if not fasta_found:
            raise ValueError(f"Failed to find .fna file with {refseq_acc} in {refseq_zip}.")
        if not gbff_found:
            raise ValueError(f"Failed to find .gbff file in {refseq_zip}.")
    # Search for segment in refseq.gbff's record for refseq_acc
    segment = None
    with open(gbff_path, 'r') as gbff_in:
        for record in SeqIO.parse(gbff_in, 'genbank'):
            if record.id == refseq_acc:
                for feature in record.features:
                    if "segment" in feature.qualifiers:
                        segment = normalize_segment(feature.qualifiers["segment"][0])
                        print(f"Found segment {segment} for {refseq_acc} in {gbff_path}")
                        break
                break
    finish_timing(start_time)
    return fasta_path, gbff_path, length, segment


def scan_ncbi_virus_metadata(ncbi_virus_metadata):
    """Get just a couple attributes that are useful for filtering sequences: length and segment."""
    acc_to_length_segment = {}
    with open(ncbi_virus_metadata, 'r') as csvfile:
        reader = csv.DictReader(csvfile)
        for row in reader:
            accession = row.get("accession")
            length = int(row.get("length", 0))
            segment = normalize_segment(row.get("segment", ""))
            acc_to_length_segment[accession] = (length, segment)
    return acc_to_length_segment


def passes_seq_filter(record, length, max_N_proportion):
    """Fasta record passes the filter if it has at most max_N_proportion of Ns"""
    return record.seq.count('N') / length <= max_N_proportion


def filter_genbank_sequences(fasta_in, fasta_out, min_length, max_N_proportion, acc_to_length_segment, refseq_segment, fasta_out_path):
    # Filter fasta file and chop descriptions to get accession as name in nextclade
    start_time = start_timing(f"Filtering GenBank sequences by length >= {min_length} and proportion of Ns <= {max_N_proportion}...")
    record_count = 0
    passed_count = 0
    for record in SeqIO.parse(fasta_in, 'fasta'):
        record_count += 1
        (length, segment) = acc_to_length_segment.get(record.id, (0, None))
        if length == 0:
            length = len(record.seq)
        # Skip if segment is given but is different from refseq_segment or sequence is shorter than min_length
        if length < min_length or (segment and segment != refseq_segment):
            continue
        # Filter by proportion of ambiguous characters
        if not passes_seq_filter(record, length, max_N_proportion):
            continue
        passed_count += 1
        # Chop sequence name after first space to limit to accession only
        record.description = record.id
        SeqIO.write(record, fasta_out, 'fasta')
    finish_timing(start_time)
    print(f"Processed {record_count} sequences from GenBank and wrote {passed_count} sequences passing filters to {fasta_out_path}.")
    return record_count, passed_count


def unpack_genbank_zip(genbank_zip, acc_to_length_segment, min_length, max_N_proportion, refseq_segment):
    """Process the fasta and data_report.jsonl files from the GenBank zip,
    filtering the fasta by length and proportion of ambiguous characters and
    keeping only the accession part of sequence names, and extracting basic
    metadata from data_report.jsonl into much more compact TSV.
    Compress fasta with lzma (xz) and TSV with gzip"""
    fasta_out_path = 'genbank.fasta.xz'
    with zipfile.ZipFile(genbank_zip, 'r') as zip_ref:
        namelist = zip_ref.namelist()
        genomic_fna_paths = [name for name in namelist if name.endswith('genomic.fna')]
        if len(genomic_fna_paths) != 1:
            raise ValueError(f"Expected exactly one genomic.fna file in {genbank_zip}, found: {genomic_fna_paths}")
        with io.TextIOWrapper(zip_ref.open(genomic_fna_paths[0], 'r'), encoding='utf-8') as fasta_in, lzma.open(fasta_out_path, 'wt') as fasta_out:
            genbank_count, passed_count = filter_genbank_sequences(fasta_in, fasta_out, min_length, max_N_proportion, acc_to_length_segment, refseq_segment, fasta_out_path)
    return fasta_out_path, genbank_count, passed_count


def open_maybe_decompress(filename, mode='rt', encoding='utf-8'):
    """
    Open a file, automatically using gzip or lzma decompression based on the file extension.
    Supports .gz, .xz, and uncompressed files.
    """
    if filename.endswith('.gz'):
        return gzip.open(filename, mode, encoding=encoding)
    elif filename.endswith('.xz'):
        return lzma.open(filename, mode, encoding=encoding)
    else:
        return open(filename, mode, encoding=encoding)


def record_to_fasta_bytes(record):
    """Convert a SeqRecord to FASTA-encoded bytes (utf-8)"""
    buf = io.StringIO()
    SeqIO.write(record, buf, 'fasta')
    return buf.getvalue().encode('utf-8')


def sanitize_name(name):
    """Replace characters that could cause trouble, such as spaces or Newick special characters,
    with underscores."""
    for c in ['[', ']', '(', ')', ':', ';', ',', "'", ' ']:
        name = name.replace(c, '_')
    return name


def align_sequences(refseq_fasta, extra_fasta, genbank_fasta, refseq_acc, min_length, max_N_proportion):
    """Run nextclade to align the filtered sequences to the reference, and pipe its output to faToVcf and gzip."""
    msa_vcf_gz = 'msa.vcf.gz'
    nextclade_err_txt = 'nextclade.align.err.log'
    start_time = start_timing(f"Aligning sequences with nextclade and converting to VCF, writing to {msa_vcf_gz}...")
    # Set up the pipeline: | nextclade | faToVcf | gzip > msa.vcf.gz
    nextclade_cmd = [
        'nextclade', 'run', '--input-ref', refseq_fasta, '--include-reference', 'true', '--output-fasta', '/dev/stdout'
    ]
    fatovcf_cmd = ['faToVcf', '-includeNoAltN', '-ref=' + refseq_acc, 'stdin', 'stdout']
    with gzip.open(msa_vcf_gz, 'wb') as vcf_out, open(nextclade_err_txt, 'wb') as nextclade_stderr:
        try:
            # Start nextclade
            nextclade_proc = subprocess.Popen(nextclade_cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=nextclade_stderr)
            # Start faToVcf, reading from nextclade's stdout
            fatovcf_proc = subprocess.Popen(fatovcf_cmd, stdin=nextclade_proc.stdout, stdout=subprocess.PIPE)
            nextclade_proc.stdout.close()  # Allow nextclade to receive SIGPIPE if faToVcf exits
            # Send fasta input(s) to nextclade_proc's stdin.  If extra_fasta duplicates sequences
            # also found in genbank_fasta, prefer the ones in extra_fasta.
            extra_ids = dict()
            if extra_fasta:
                with open_maybe_decompress(extra_fasta) as ef:
                    for record in SeqIO.parse(ef, 'fasta'):
                        record.description = sanitize_name(record.description)
                        record.id = record.description
                        if record.id in extra_ids:
                            print(f"Fasta ID '{record.id}' has already been used in {extra_fasta}, discarding duplicate.", file=sys.stderr)
                        else:
                            extra_ids[record.id] = True
                            length = len(record.seq)
                            if length < min_length or not passes_seq_filter(record, length, max_N_proportion):
                                continue
                            nextclade_proc.stdin.write(record_to_fasta_bytes(record))
            # genbank_fasta has already been filtered for length & Ns, but discard duplicates
            with open_maybe_decompress(genbank_fasta) as gf:
                for record in SeqIO.parse(gf, 'fasta'):
                    if record.id in extra_ids:
                        print(f"GenBank ID '{record.id}' was already used in {extra_fasta}, discarding duplicate.", file=sys.stderr)
                    else:
                        nextclade_proc.stdin.write(record_to_fasta_bytes(record))
            nextclade_proc.stdin.close()
            # Write faToVcf's output to gzip file
            for chunk in iter(lambda: fatovcf_proc.stdout.read(8192), b''):
                vcf_out.write(chunk)
            fatovcf_proc.stdout.close()
            fatovcf_proc.wait()
            nextclade_proc.wait()
        except subprocess.CalledProcessError as e:
            print(f"nextclade | faToVcf failed: {e.stderr}", file=sys.stderr)
            raise
    finish_timing(start_time)
    start_time = start_timing(f"Compressing {nextclade_err_txt}...")
    # gzip nextclade.align.err.log (note: using gzip.open above did not work, it was written uncompressed.)
    run_command(['gzip', '-f', nextclade_err_txt])
    # Count the number of samples in the #CHROM header line of msa.vcf.gz
    aligned_count = 0
    with gzip.open(msa_vcf_gz, 'rt') as vcf_in:
        for line in vcf_in:
            if line.startswith('#CHROM'):
                header = line.strip().split('\t')
                aligned_count = len(header) - 9  # Subtract fixed columns
                break
    finish_timing(start_time)
    return msa_vcf_gz, aligned_count


def make_empty_tree():
    """Create an empty tree file to be used as a starting point for usher-sampled."""
    empty_tree_path = 'empty_tree.nwk'
    with open(empty_tree_path, 'w') as f:
        f.write("()\n")
    return empty_tree_path


def run_usher_sampled(tree, vcf):
    """Build the tree.  Use docker --platform linux/amd64 so this will work even on Mac with ARM CPU. """
    pb_out = 'usher_sampled.pb.gz'
    start_time = start_timing(f"Running usher-sampled on {vcf}...")
    tree_flag = "-t" if tree.endswith('.nwk') else "-i"
    command = ['usher-sampled', '-A', '-e', '5', tree_flag, tree, '-v', vcf, '-o', pb_out,
               '--optimization_radius', '0', '--batch_size_per_process', '100']
    run_command(command, stdout_filename='usher-sampled.out.log', stderr_filename='usher-sampled.err.log')
    finish_timing(start_time)
    return pb_out


def run_matoptimize(pb_file, vcf_file, update):
    """Run matOptimize to clean up after usher-sampled"""
    pb_out = 'optimized.unfiltered.pb.gz'
    start_time = start_timing(f"Running matOptimize on {pb_file}...")
    command_no_vcf = ['matOptimize', '-m', '0.00000001', '-M', '1', '-i', pb_file, '-o', pb_out]
    # Unless we're doing an update, in which case the VCF covers only the new sequences,
    # first try with VCF, which is less tested but should give better results because it includes
    # info about which bases are ambiguous or N.  That info is lost when usher imputes values.
    if update:
        run_command(command_no_vcf, stdout_filename='matOptimize.out.log', stderr_filename='matOptimize.err.log')
    else:
        command_vcf = command_no_vcf.copy()
        command_vcf.append("-v")
        command_vcf.append(vcf_file)
        if not run_command(command_vcf, stdout_filename='matOptimize.out.log', stderr_filename='matOptimize.err.log', fail_ok=True):
            finish_timing(start_time)
            start_time = start_timing("matOptimize with VCF failed, trying again without VCF...")
            run_command(command_no_vcf, stdout_filename='matOptimize.out.log', stderr_filename='matOptimize.err.log')
    finish_timing(start_time)
    return pb_out


def run_matutils_filter(opt_unfiltered_tree, max_parsimony, max_branch_length):
    """Run matUtils extract to filter sequences/branches and collapse tree post-matOptimize"""
    pb_out = update_tree_input
    sample_names_out = 'tree_samples.txt'
    start_time = start_timing(f"Running matUtils to filter (--max-parsimony {max_parsimony} --max-branch-length {max_branch_length})...")
    command = ['matUtils', 'extract', '-i', opt_unfiltered_tree,
               '--max-parsimony', str(max_parsimony),
               '--max-branch-length', str(max_branch_length),
               '--collapse-tree',
               '--used-samples', sample_names_out,
               '-o', pb_out]
    run_command(command, stdout_filename='matUtils.filter.out.log', stderr_filename='matUtils.filter.err.log')
    finish_timing(start_time)
    # Run matUtils summary to get the final number of samples in the tree
    start_time = start_timing("Counting samples in the final tree...")
    command = ['matUtils', 'summary', '-i', pb_out]
    summary_output = run_command_get_stdout(command)
    for line in summary_output.split('\n'):
        if line.startswith('Total Samples in Tree:'):
            tree_tip_count = int(line.split(':')[1].strip())
            break
    finish_timing(start_time)
    return pb_out, sample_names_out, tree_tip_count


def get_existing_nextclade_assignments(update, starting_tree_accessions, nextclade_path):
    """Get existing Nextclade assignments, and the number of columns to expect, from the old TSV file about to be replaced."""
    if not update or not nextclade_path:
        return {}, None
    existing_assignments = {}
    nextclade_tsv = update_nextclade_input if os.path.exists(update_nextclade_input) else update_nextclade_input + ".gz"
    existing_column_count = None
    if os.path.exists(nextclade_tsv):
        with open_maybe_decompress(nextclade_tsv) as tsv_in:
            header_cols = tsv_in.readline().rstrip('\n').split('\t')
            # 'dynamic' might have added clade and/or phenotype, motif etc. columns; keep all of them
            col_count = len(header_cols) - 1
            if existing_column_count is None:
                existing_column_count = col_count
            elif existing_column_count != col_count:
                raise ValueError(f"Inconsistent Nextclade column count in {nextclade_tsv}: {existing_column_count} vs {col_count}")
            for line in tsv_in:
                fields = line.rstrip('\n').split('\t')
                seq_name = fields[0]
                clades = fields[1:]
                existing_assignments[seq_name] = clades
    else:
        print(f"Nextclade TSV file {nextclade_tsv} not found. Only new items will have clade assignments.", file=sys.stderr)
    return existing_assignments, existing_column_count


def run_nextclade(nextclade_path, nextclade_clade_columns, existing_nextclade_assignments, existing_column_count, genbank_fasta, extra_fasta, refseq_fasta):
    """Run nextclade to assign sequences to clades.  Return dict mapping accession to clade."""
    if not nextclade_path:
        return {}, ""
    nextclade_tsv = update_nextclade_input
    # Nextclade --output-columns-selection does not accept individual customClades names; instead use 'dynamic'
    # to get all of them.  See https://github.com/nextstrain/nextclade/issues/1667 .
    nextclade_col_list = nextclade_clade_columns.split(',')
    if len(nextclade_col_list) > 1:
        output_columns = 'seqName,clade,dynamic'
    else:
        output_columns = 'seqName,clade'
    start_time = start_timing(f"Running nextclade to assign clades (dataset: {nextclade_path})...")
    command = ['nextclade', 'run',
               '--dataset-name', nextclade_path,
               '--output-columns-selection', output_columns,
               '--output-tsv', nextclade_tsv,
               genbank_fasta, refseq_fasta]
    if extra_fasta:
        command.append(extra_fasta)
    run_command(command, stdout_filename='nextclade.clade.out.log', stderr_filename='nextclade.clade.err.log')
    finish_timing(start_time)
    # Read the TSV file and return a dict mapping sequence names to clades
    start_time = start_timing(f"Reading nextclade assignments from {nextclade_tsv} to add to metadata...")
    new_assignments = {}
    with open(nextclade_tsv, 'r') as tsv_in:
        header_cols = tsv_in.readline().rstrip('\n').split('\t')
        # 'dynamic' might have added clade and/or phenotype, motif etc. columns; keep all of them
        nextclade_clade_columns = ','.join(header_cols[1:])
        nextclade_column_count = len(header_cols) - 1
        if existing_column_count is not None and nextclade_column_count != existing_column_count:
            print(f"Existing nextclade annotations had {existing_column_count} columns but new annotations have {nextclade_column_count} columns ({header_cols[1:]})." +
                  f"\nRemove {update_nextclade_input} and run again, to rebuild with the latest nextclade dataset '{nextclade_path}'", file=sys.stderr)
            sys.exit(1)
        for line in tsv_in:
            fields = line.rstrip('\n').split('\t')
            seq_name = fields[0]
            clades = fields[1:]
            new_assignments[seq_name] = clades
    if existing_column_count is not None:
        nextclade_assignments = existing_nextclade_assignments
        nextclade_assignments.update(new_assignments)
    else:
        nextclade_assignments = new_assignments
    finish_timing(start_time)
    if existing_column_count is not None:
        # Append the existing assignments to the new little update nextclade.clade.tsv so they're not lost.
        with open(nextclade_tsv, 'a') as tsv_out:
            for acc, clade_columns in existing_nextclade_assignments.items():
                if acc not in new_assignments:
                    print(acc + "\t" + "\t".join(clade_columns), file=tsv_out)
    return nextclade_assignments, nextclade_clade_columns


def get_numerical_date(date):
    """Date should be YYYY-MM-DD or YYYY-MM or YYYY.  Return year.fraction"""
    if not date:
        return 0.0
    groups = re.match('^([0-9]{4})(-([0-9]{2})(-([0-9]{2}))?)?$', date)
    if not groups:
        return 0.0
    year = int(groups[1])
    if groups[3]:
        month = int(groups[3])
        day = int(groups[5]) if groups[5] else 15
    else:
        month = 7
        day = 1
    date_obj = datetime.date(year, month, day)
    day_of_year = date_obj.timetuple().tm_yday
    fraction = day_of_year / 365.25
    return year + fraction


def fudge_isolate(isolate, accession, date, country, year):
    """If isolate looks like a full /-separated descriptor then use it.
    Otherwise try to approximate it from available metadata bits and pieces."""
    full = None
    if '/' in isolate:
        full = isolate
    else:
        if country and isolate and year:
            full = '/'.join([country, isolate, year])
        elif country and year:
            full = '/'.join([country, year])
        elif isolate and year:
            full = '/'.join([isolate, year])
        elif isolate:
            full = isolate
    return full


def make_display_name(isolate, strain, date, accession, country):
    """Make a display name with whatever info is available in isolate or strain, the country and date in addition to accession."""
    if not isolate:
        isolate = strain
    elif strain and len(isolate) < len(strain):
        # Sometimes people use nice strain like "MVi/Marseille.FRA/21.19/20[D8]" but goofy isolate
        # like "9061923678/9051834309" for MZ031240.1.  If strain is longer, use it.
        isolate = strain
    groups = re.match('^([0-9]{4})(-[0-9]{2})?(-[0-9]{2})?$', date)
    year = None
    if groups:
        year = groups[1]
    if not date:
        date = '?'
    full = fudge_isolate(isolate, accession, date, country, year)
    if full:
        name = '|'.join([full, accession, date])
    else:
        name = '|'.join([accession, date])
    return name


def get_nextclade_column_list(nextclade_clade_columns):
    if not nextclade_clade_columns:
        return []
    nextclade_col_list = nextclade_clade_columns.split(',')
    nextclade_col_list = [col if col.startswith('nextclade') or col.startswith('Nextclade') else 'nextclade_' + col for col in nextclade_col_list]
    return nextclade_col_list


def finalize_metadata(ncbi_virus_metadata, nextclade_assignments, nextclade_clade_columns, refseq_segment, sample_names):
    """Make a file for renaming sequence names in tree to include country, isolate name and date when available.
    Prepare metadata for taxonium."""
    rename_out = "rename.tsv"
    metadata_out = "metadata.tsv.gz"
    start_time = start_timing(f"Finalizing sequence names for display and metadata, writing to {metadata_out}...")
    remove_segment = not refseq_segment
    date_min = None
    date_max = None
    # Use a subprocess to get input from grep -Fwf {sample_names} {ncbi_virus_metadata}
    grep_cmd = ['grep', '-Fwf', sample_names, ncbi_virus_metadata]
    with subprocess.Popen(grep_cmd, stdout=subprocess.PIPE, text=True) as grep_proc, \
            open(rename_out, 'w') as r_out, gzip.open(metadata_out, 'wt') as m_out:
        # Read header from ncbi_virus_metadata (not filtered by grep)
        with open(ncbi_virus_metadata, 'rt') as csv_in:
            header = csv_in.readline().rstrip('\n').split(',')
        accession_idx = header.index('accession')
        isolate_idx = header.index('isolate')
        strain_idx = header.index('strain')
        date_idx = header.index('date')
        country_loc_idx = header.index('country_location')
        # In metadata header, make separate columns for country and location:
        header_mod = header[:country_loc_idx] + ['country', 'location'] + header[country_loc_idx+1:]
        # Rename 'strain' to 'gb_strain' because I add 'strain' at the beginning below, following the
        # nextstrain convention of 'strain' meaning the full descriptor
        header_mod[header_mod.index('strain')] = 'gb_strain'
        # Remove 'segment' if refseq has no segment
        segment_idx = header_mod.index('segment')
        if remove_segment:
            header_mod.pop(segment_idx)
        # Add numerical date column
        header_mod.append('num_date')
        # No header for rename_out; write header for metadata_out
        metadata_header = 'strain\t' + '\t'.join(header_mod)
        if nextclade_clade_columns:
            nextclade_col_list = get_nextclade_column_list(nextclade_clade_columns)
            metadata_header += '\t' + '\t'.join(nextclade_col_list)
        m_out.write(metadata_header + '\n')

        # Process filtered lines from grep
        reader = csv.reader(grep_proc.stdout, delimiter=',')
        for row in reader:
            accession = row[accession_idx].strip()
            isolate = sanitize_name(row[isolate_idx].strip())
            strain = sanitize_name(row[strain_idx].strip())
            date = row[date_idx].strip()
            country_loc = row[country_loc_idx].strip()
            # Make separate columns for country and location
            country, location = country_loc.split(':', 1) if ':' in country_loc else (country_loc, '')
            country = sanitize_name(country.strip())
            location = location.strip()
            row_mod = row[:country_loc_idx] + [country, location] + row[country_loc_idx+1:]
            if remove_segment:
                row_mod.pop(segment_idx)
            # Add numerical date column (year.fraction)
            if date and re.match('^[0-9]{4}(-[0-9]{2}(-[0-9]{2})?)?$', date):
                num_date = get_numerical_date(date)
                if date_min is None or num_date < date_min:
                    date_min = num_date
                if date_max is None or num_date > date_max:
                    date_max = num_date
                row_mod.append(f"{num_date:.6f}")
            else:
                row_mod.append("")
            name = make_display_name(isolate, strain, date, accession, country)
            metadata = name + '\t' + '\t'.join(row_mod)
            if nextclade_clade_columns:
                clades = nextclade_assignments.get(accession, [''] * len(nextclade_col_list))
                metadata += '\t' + '\t'.join(clades)
            # Write output to both renaming file and metadata file
            r_out.write("\t".join([accession, name]) + '\n')
            m_out.write(metadata + '\n')
    finish_timing(start_time)
    return metadata_out, rename_out, date_min, date_max


def rename_seqs(pb_in, rename_tsv):
    pb_out = "viz.pb.gz"
    start_time = start_timing(f"Renaming sequences in {pb_in} to make {pb_out}...")
    command = ['matUtils', 'mask', '-i', pb_in, '--rename-samples', rename_tsv, '-o', pb_out]
    run_command(command, stdout_filename='matUtils.rename.out.log', stderr_filename='matUtils.rename.err.log')
    finish_timing(start_time)
    return pb_out


def dump_newick(pb_in):
    """Dump the phylogenetic tree in Newick format."""
    newick_out = "viz.nwk"
    start_time = start_timing(f"Writing tree in Newick format to {newick_out}.gz...")
    command = ['matUtils', 'extract', '-i', pb_in, '--write-tree', newick_out]
    run_command(command, stdout_filename='/dev/null', stderr_filename='/dev/null')
    command = ['gzip', '-f', newick_out]
    run_command(command, stdout_filename='/dev/null', stderr_filename='/dev/null')
    finish_timing(start_time)


def get_header(tsv_in):
    with gzip.open(tsv_in, 'rt') as tsv:
        header = tsv.readline().split('\t')
        for idx, field in enumerate(header):
            header[idx] = field.strip()
    return header


def make_taxonium_config(date_min, date_max, nextclade_clade_columns):
    """Make a config file with a color gradient from date_min to date_max."""
    config_out = "taxonium_config.json"
    color_by_options = ["meta_country", "meta_location", "meta_num_date", "meta_host", "meta_serotype"]
    if nextclade_clade_columns:
        color_by_options += ["meta_" + col for col in get_nextclade_column_list(nextclade_clade_columns)]
    color_by_options += ["genotype", "None"]
    config = {"colorRamps": {"meta_num_date": {"scale": [[date_min, "#0000F8"], [date_max, "#F80000"]]}},
              "customNames": {"meta_num_date": "Date (numeric)",
                              "meta_gb_strain": "Strain"},
              "colorBy": {"colorByOptions": color_by_options}}
    with open(config_out, 'w') as f:
        json.dump(config, f)
    return config_out


def usher_to_taxonium(pb_in, metadata_in, refseq_gbff, tip_count, species, refseq_acc, date_min, date_max, nextclade_clade_columns):
    jsonl_out = "tree.jsonl.gz"
    start_time = start_timing(f"Running usher_to_taxonium to make {jsonl_out}...")
    columns = ','.join(get_header(metadata_in))
    title = f"{tip_count} {species} sequences from GenBank aligned to {refseq_acc}"
    config = make_taxonium_config(date_min, date_max, nextclade_clade_columns)
    command = ['usher_to_taxonium', '--input', pb_in, '--metadata', metadata_in,
               '--columns', columns, '--title', title, '--config_json', config,
               '--genbank', refseq_gbff, '--output', jsonl_out]
    if not run_command(command, stdout_filename='utt.out.log', stderr_filename='utt.err.log', fail_ok=True):
        finish_timing(start_time)
        start_time = start_timing("usher_to_taxonium failed with --genbank, trying again without --genbank...")
        command = ['usher_to_taxonium', '--input', pb_in, '--metadata', metadata_in,
                   '--columns', columns, '--title', title, '--config_json', config,
                   '--output', jsonl_out]
        run_command(command, stdout_filename='utt.out.log', stderr_filename='utt.err.log')
    finish_timing(start_time)
    return jsonl_out


def write_output_stats(refseq_acc, refseq_length, gb_count, filtered_count, aligned_count, tree_tip_count):
    output_stats = [["refseq_acc", "refseq_length", "gb_count", "filtered_count", "aligned_count", "tree_tip_count"],
                    [refseq_acc, refseq_length, gb_count, filtered_count, aligned_count, tree_tip_count]]
    print("Writing output stats to output_stats.tsv.")
    with open("output_stats.tsv", "w") as f:
        for row in output_stats:
            f.write("\t".join(map(str, row)) + "\n")


def get_tree_names(pb_in):
    """Return a set containing the names of the tips in pb_in."""
    tree_names = set()
    start_time = start_timing(f"Extracting tree names from {pb_in}...")
    command = ["matUtils", "extract", "-i", pb_in, "--used-samples", "tree_samples.txt"]
    run_command(command, stdout_filename="/dev/null", stderr_filename="/dev/null")
    with open("tree_samples.txt", "r") as f:
        for line in f:
            tree_names.add(line.strip())
    finish_timing(start_time)
    return tree_names


def find_new_accessions(tree_names, acc_to_length_segment, min_length, refseq_segment):
    """Return a list of GenBank accessions in acc_to_length_segment that are not in the current tree and probably wouldn't be filtered later."""
    all_accessions = set(acc_to_length_segment.keys())
    # Exclude accessions whose metadata has a segment that is different from RefSeq
    exclude_accessions = {acc for acc, (length, seg) in acc_to_length_segment.items() if length < min_length or seg and seg != refseq_segment}
    new_accessions = all_accessions - tree_names - exclude_accessions
    return list(new_accessions)


def get_new_extra_fasta(tree_names, extra_fasta):
    """Get a new extra_fasta file containing only sequences not already in the tree."""
    if not extra_fasta:
        return ""
    now = datetime.datetime.now()
    new_extra_fasta = "new_extra." + now.strftime("%Y-%m-%d_%H-%M-%S") + ".fasta"
    start_time = start_timing(f"Filtering {extra_fasta} to get new sequences not already in tree, writing to {new_extra_fasta}...")
    new_count = 0
    with open_maybe_decompress(extra_fasta) as ef, open(new_extra_fasta, 'w') as nef:
        for record in SeqIO.parse(ef, 'fasta'):
            record.description = sanitize_name(record.description)
            record.id = record.description
            if record.id not in tree_names:
                SeqIO.write(record, nef, 'fasta')
                new_count += 1
    finish_timing(start_time)
    if new_count == 0:
        print(f"No new sequences found in {extra_fasta}, skipping it.")
        return ""
    else:
        print(f"Wrote {new_count} new sequences to {new_extra_fasta}.")
        return new_extra_fasta


def run_nextclade_on_existing(ncbi, starting_tree_accessions, nextclade_path, nextclade_clade_columns,
                              extra_fasta, refseq_fasta):
    """Download GenBank sequences that are already in the tree.  Run nextclade on those plus extra_fasta."""
    old_genbank_zip = "genbank_old.zip"
    start_time = start_timing("Downloading and unpacking GenBank genomes that were already in the tree, for nextclade...")
    ncbi.download_genbank_accessions(list(starting_tree_accessions), old_genbank_zip)
    old_genbank_fasta, _, _ = unpack_genbank_zip(old_genbank_zip, {}, 0, 1.0, "")
    finish_timing(start_time)
    print("Running nextclade on genomes that were already in the tree.")
    run_nextclade(nextclade_path, nextclade_clade_columns, {}, None, old_genbank_fasta, extra_fasta, refseq_fasta)
    os.remove(old_genbank_zip)
    if not os.path.exists(update_nextclade_input):
        print(f"Expected {update_nextclade_input} to be created but it's not found.", sys.stderr)
        sys.exit(1)


def main():
    parser = argparse.ArgumentParser(prog="viral_usher_build")
    parser.add_argument("--config", type=str, required=True, help="Path to config file input")
    parser.add_argument("--update", action="store_true",
                        help="Add only new sequences to existing tree instead of building tree from scratch")

    args = parser.parse_args()

    print(f"Running pipeline with config file: {args.config}")
    config_contents = config.parse_config(args.config)
    refseq_acc = config_contents['refseq_acc']
    assembly_id = config_contents['refseq_assembly']
    taxid = config_contents['taxonomy_id']
    nextclade_path = config_contents.get('nextclade_dataset', '')
    nextclade_clade_columns = config_contents.get('nextclade_clade_columns', 'clade')
    min_length_proportion = float(config_contents.get('min_length_proportion', config.DEFAULT_MIN_LENGTH_PROPORTION))
    max_N_proportion = float(config_contents.get('max_N_proportion', config.DEFAULT_MAX_N_PROPORTION))
    max_parsimony = int(config_contents.get('max_parsimony', str(config.DEFAULT_MAX_PARSIMONY)))
    max_branch_length = int(config_contents.get('max_branch_length', str(config.DEFAULT_MAX_BRANCH_LENGTH)))
    extra_fasta = config_contents.get('extra_fasta', '')
    species = config_contents.get('species', None)
    if not species:
        start_time = start_timing(f"Looking up species name for RefSeq {refseq_acc}...")
        species, _ = ncbi_helper.NcbiHelper().get_species_taxid_for_refseq(refseq_acc)
        finish_timing(start_time)
    refseq_zip = f"{refseq_acc}.zip"
    genbank_zip = f"genbank_{taxid}.zip"

    # Download the RefSeq genome and annotations
    ncbi = ncbi_helper.NcbiHelper()
    start_time = start_timing(f"Downloading RefSeq {refseq_acc} (Assembly {assembly_id}) genome to {refseq_zip}...")
    ncbi.download_refseq(assembly_id, refseq_zip)
    finish_timing(start_time)
    refseq_fasta, refseq_gbff, refseq_length, refseq_segment = unpack_refseq_zip(refseq_zip, refseq_acc)
    min_length = int(refseq_length * min_length_proportion)

    starting_tree_accessions = []
    if (args.update):
        # Make sure we have required inputs
        if not os.path.exists(update_tree_input):
            print(f"No existing {update_tree_input} found, cannot do update.  Run without --update to build tree from scratch.", file=sys.stderr)
            sys.exit(1)
        starting_tree_accessions = get_tree_names(update_tree_input)
        if nextclade_path and not os.path.exists(update_nextclade_input) and not os.path.exists(update_nextclade_input + ".gz"):
            print(f"No existing {update_nextclade_input} found; recreating it.")
            run_nextclade_on_existing(ncbi, starting_tree_accessions, nextclade_path, nextclade_clade_columns,
                                      extra_fasta, refseq_fasta)

    # Get metadata from NCBI Virus API
    start_time = start_timing(f"Querying NCBI Virus API for metadata for all GenBank sequences for taxid {taxid}...")
    ncbi_virus_metadata = "ncbi_virus_metadata.csv"
    ncbi.query_ncbi_virus_metadata(taxid, ncbi_virus_metadata)
    acc_to_length_segment = scan_ncbi_virus_metadata(ncbi_virus_metadata)
    finish_timing(start_time)

    # Download GenBank genomes for the given Taxonomy ID: only the new ones if --update, otherwise all of them
    got_genbank = False
    if (args.update):
        new_accessions = find_new_accessions(starting_tree_accessions, acc_to_length_segment, min_length, refseq_segment)
        if len(new_accessions) > 0:
            start_time = start_timing(f"Downloading {len(new_accessions)} new GenBank genomes for taxid {taxid} to {genbank_zip}...")
            ncbi.download_genbank_accessions(new_accessions, genbank_zip)
            got_genbank = True
            finish_timing(start_time)
        else:
            print("No new GenBank accessions found, skipping GenBank download.")
        extra_fasta = get_new_extra_fasta(starting_tree_accessions, extra_fasta)
    else:
        start_time = start_timing(f"Downloading all GenBank genomes for taxid {taxid} to {genbank_zip}...")
        ncbi.download_genbank(taxid, genbank_zip)
        got_genbank = True
        finish_timing(start_time)
    if got_genbank:
        genbank_fasta, gb_count, filtered_count = unpack_genbank_zip(genbank_zip, acc_to_length_segment, min_length, max_N_proportion, refseq_segment)
    else:
        genbank_fasta = "/dev/null"
        gb_count = 0
        filtered_count = 0

    # The core of the pipeline: align sequences, build tree, finalize metadata
    msa_vcf, aligned_count = align_sequences(refseq_fasta, extra_fasta, genbank_fasta, refseq_acc, min_length, max_N_proportion)
    if args.update:
        if aligned_count == 0:
            preopt_tree = "usher_sampled.pb.gz"
            shutil.copy(update_tree_input, preopt_tree)
        else:
            preopt_tree = run_usher_sampled(update_tree_input, msa_vcf)
    else:
        empty_tree = make_empty_tree()
        preopt_tree = run_usher_sampled(empty_tree, msa_vcf)
    opt_unfiltered_tree = run_matoptimize(preopt_tree, msa_vcf, args.update)
    opt_tree, sample_names, tree_tip_count = run_matutils_filter(opt_unfiltered_tree, max_parsimony, max_branch_length)
    existing_nextclade_assignments, existing_column_count = get_existing_nextclade_assignments(args.update, starting_tree_accessions, nextclade_path)
    nextclade_assignments, nextclade_clade_columns = run_nextclade(nextclade_path, nextclade_clade_columns,
                                                                   existing_nextclade_assignments, existing_column_count,
                                                                   genbank_fasta, extra_fasta, refseq_fasta)
    metadata_tsv, rename_tsv, date_min, date_max = finalize_metadata(ncbi_virus_metadata, nextclade_assignments, nextclade_clade_columns,
                                                                     refseq_segment, sample_names)
    viz_tree = rename_seqs(opt_tree, rename_tsv)
    dump_newick(viz_tree)
    write_output_stats(refseq_acc, refseq_length, gb_count, filtered_count, aligned_count, tree_tip_count)
    usher_to_taxonium(viz_tree, metadata_tsv, refseq_gbff, tree_tip_count, species, refseq_acc, date_min, date_max, nextclade_clade_columns)


if __name__ == "__main__":
    main()



import requests
import json
import sys


def nextclade_get_index():
    """Get Nextclade's v3/index.json file and extract the parts useful for matching the
    user's search term and listing available clade annotations.  If the request fails,
    warn and return an empty list."""
    url = "https://data.clades.nextstrain.org/v3/index.json"
    try:
        response = requests.get(url)
        response.raise_for_status()
    except requests.RequestException as e:
        print(f"Error fetching Nextclade index: {e}", file=sys.stderr)
        return []

    try:
        index_data = response.json()
    except json.JSONDecodeError:
        print("Error decoding JSON from Nextclade index.", file=sys.stderr)
        return []

    entries = []

    for collection in index_data.get('collections', []):
        for dataset in collection.get('datasets', []):
            cladeCount = dataset.get('capabilities', {}).get('clades', 0)
            # Skip entries without clade annotations
            if cladeCount > 0:
                entry = {}
                entry["path"] = dataset.get('path')
                entry["name"] = dataset.get('attributes', {}).get('name')
                entry["segment"] = dataset.get('attributes', {}).get('segment')
                entry["clades"] = {"clade": cladeCount}
                # If customClades are included, break out their counts too
                customClades = dataset.get('capabilities', {}).get('customClades', {})
                if customClades:
                    for customClade, count in customClades.items():
                        entry["clades"][customClade] = count
                entries.append(entry)

    return entries



"""Get data from NCBI using APIs (EUtils, Virus, Datasets) or the datasets command-line tool."""

import requests
import logging
import sys
import time
import json
import subprocess
import xml.etree.ElementTree as ET

NCBI_DATASETS_V2_BASE = "https://api.ncbi.nlm.nih.gov/datasets/v2"
NCBI_DATASETS_TAXONOMY_SUGGEST_URL = f"{NCBI_DATASETS_V2_BASE}/taxonomy/taxon_suggest"
NCBI_EUTILS_BASE = "https://eutils.ncbi.nlm.nih.gov/entrez/eutils"
NCBI_EUTILS_SEARCH_URL = f"{NCBI_EUTILS_BASE}/esearch.fcgi"
NCBI_EUTILS_SUMMARY_URL = f"{NCBI_EUTILS_BASE}/esummary.fcgi"
NCBI_EUTILS_ELINK_URL = f"{NCBI_EUTILS_BASE}/elink.fcgi"
NCBI_EUTILS_EFETCH_URL = f"{NCBI_EUTILS_BASE}/efetch.fcgi"


class NcbiHelper:
    max_retries = 5
    retry_delay = 60
    retry_delay_factor = 3

    def __init__(self):
        self.logger = logging.getLogger(__name__)

    def set_debug(self, debug):
        if debug:
            self.logger.setLevel(logging.DEBUG)
        else:
            self.logger.setLevel(logging.INFO)

    def run_command(self, command):
        """Run a command using subprocess.run, check for errors."""
        self.logger.debug(f"Running command: {' '.join(command)}")
        try:
            result = subprocess.run(command, check=True, capture_output=True, text=True)
            self.logger.debug(f"Command output: {result.stdout}")
        except subprocess.CalledProcessError as e:
            self.logger.error(f"Command '{' '.join(command)}' failed with exit code {e.returncode}")
            self.logger.error(f"stderr: {e.stderr}")
            sys.exit(1)

    def query_with_retry(self, func, *args, **kwargs):
        """Call func with args and kwargs, retrying on failure."""
        time.sleep(1)
        attempt = 0
        delay = self.retry_delay
        while attempt < self.max_retries:
            attempt += 1
            try:
                return func(*args, **kwargs)
            except Exception as e:
                self.logger.error(f"Function {func.__name__} failed on attempt {attempt} with error: {e}")
            if attempt < self.max_retries:
                time.sleep(delay)
                delay *= self.retry_delay_factor
            else:
                self.logger.error(f"Function {func.__name__} failed after {self.max_retries} attempts.")
                sys.exit(1)

    def check_response(self, resp):
        if resp.status_code != 200:
            self.logger.error(f"API request failed: {resp.status_code}")
            try:
                self.logger.error(json.dumps(resp.json(), indent=2))
            except json.JSONDecodeError:
                self.logger.error(resp.text)
            return False
        return True

    def get_request(self, url, params):
        """Send a GET request, check the response, and return the JSON data.  Retry on failure.  Sleep to rate-limit."""
        def get_request_core(url, params):
            resp = requests.get(url, params=params)
            if self.check_response(resp):
                try:
                    result = resp.json()
                    return result
                except json.JSONDecodeError:
                    raise Exception(f"Expected JSON from status 200 response but got decode error from url {url} with params:\n" +
                                    json.dumps(params, indent=2) + "\nResponse:\n" + resp.text)
            else:
                raise Exception(f"Bad response from {url} with params:\n{json.dumps(params, indent=2)}")
        return self.query_with_retry(get_request_core, url, params)

    def eutils_request(self, url, params):
        """Send a request, check the response, and return the JSON data."""
        params["retmode"] = "json"
        return self.get_request(url, params)

    def esearch_request(self, db, term):
        """Use ESearch to search for term in db.  Return a list of GI numbers."""
        data = self.eutils_request(NCBI_EUTILS_SEARCH_URL, {"db": db, "term": term})
        return data.get("esearchresult", {}).get("idlist", [])

    def esearch_request_one(self, db, term):
        """Use ESearch to search for term in db.  Warn if there is not exactly one result. Return the first GI number."""
        idlist = self.esearch_request(db, term)
        if not idlist or len(idlist) == 0:
            self.logger.warning(f"No ID found for {db} term '{term}'.")
            return None
        elif len(idlist) > 1:
            self.logger.warning(f"Multiple IDs found for {db} term '{term}'. Using the first one.")
        return idlist[0]

    def elink_request(self, db, dbfrom, id):
        """Use ELink to get links from dbfrom to db for the given ID.  Return a list of linked GI numbers.  Assumes that there is only one linksetdb."""
        data = self.eutils_request(NCBI_EUTILS_ELINK_URL, {"db": db, "dbfrom": dbfrom, "id": id})
        linksets = data.get("linksets", [])
        if not linksets:
            self.logger.warning(f"No linksets found for {dbfrom} GI {id}.  data: {data}")
            return []
        if len(linksets) > 1:
            self.logger.warning(f"Multiple linksets found for {dbfrom} GI {id}. Using the first one.  data: {data}")
        return linksets[0].get("ids")

    def elink_request_one(self, db, dbfrom, id):
        """Use ELink to get links from dbfrom to db for the given ID.  Warn if there is not exactly one result. Return the first linked GI number."""
        links = self.elink_request(db, dbfrom, id)
        if not links or len(links) == 0:
            self.logger.warning(f"No {db} links found for {dbfrom} GI {id}.")
            return None
        elif len(links) > 1:
            self.logger.warning(f"Multiple {db} links found for {dbfrom} GI {id}. Using the first one.")
        return links[0]

    def esummary_request(self, db, ids):
        """Use ESummary to get summaries for the given IDs in the specified database.  Return a dictionary of summaries."""
        data = self.eutils_request(NCBI_EUTILS_SUMMARY_URL, {"db": db, "id": ",".join(ids)})
        return data.get("result", {})

    def esummary_request_one(self, db, id):
        """Use ESummary to get a summary for the given ID in the specified database.  Return the summary dictionary for id."""
        data = self.eutils_request(NCBI_EUTILS_SUMMARY_URL, {"db": db, "id": id})
        if not data or "result" not in data or id not in data["result"]:
            self.logger.warning(f"No summary found for {db} ID {id}.")
            return None
        return data.get("result", {}).get(id, {})

    def get_taxonomy_entries(self, species_name):
        """Use the NCBI Datasets API to look up the Taxonomy ID(s) for a given species name.
        Return [ { "sci_name": ..., "tax_id": ... }, ...] for all matches."""
        url = f"{NCBI_DATASETS_TAXONOMY_SUGGEST_URL}/{species_name}"
        data = self.get_request(url, {})
        return data.get("sci_name_and_ids", [])

    def get_assembly_acc_for_refseq_acc(self, refseq_acc):
        """Use the NCBI EUtils API to look up the Assembly accession for a given RefSeq accession."""
        # First use NCBI Esearch to search for the Assembly GI number for the given RefSeq GI number
        refseq_gi = self.esearch_request_one("assembly", f"{refseq_acc}[Accession]")
        if not refseq_gi:
            return None
        # Next use NCBI ELink to get the assembly GI number for the given RefSeq GI number
        assembly_gi = self.elink_request_one("assembly", "nuccore", refseq_gi)
        if not assembly_gi:
            return None
        # Now use the assembly GI number to get the assembly accession
        return self.esummary_request_one("assembly", assembly_gi).get("assemblyaccession")

    def get_refseqs_for_taxid(self, taxid):
        """Use the NCBI EUtils API to look up RefSeq IDs for a given Taxonomy ID.
        Return [ { "accession": ..., "title": ..., "strain": ..., "uid": ... }, ...] for all matches."""
        # First use NCBI Esearch to search for RefSeq GI numbers for the given Taxonomy ID
        ids = self.esearch_request("nuccore", f"txid{taxid}[Organism:exp] AND srcdb_refseq[PROP]")
        if not ids:
            return None
        # Next use NCBI Esummary to get actual RefSeq accession numbers, titles etc for the GI numbers
        summaries = self.esummary_request("nuccore", ids)
        # Reformat the results into just what we need
        refseqs = []
        for refseq_gi in ids:
            entry = summaries.get(refseq_gi, {})
            accession = entry.get("accessionversion")
            if not accession:
                self.logger.warning(f"No accession for RefSeq GI number {refseq_gi}.")
                continue
            refseqs.append({
                "accession": accession,
                "title": entry.get("title", "No title"),
                "strain": entry.get("strain", "No strain"),
                "uid": refseq_gi
            })
        return refseqs

    def get_species_taxid_for_refseq(self, refseq_id):
        """Look up the Taxonomy ID for the given RefSeq ID using the NCBI EUtils API and Assembly database."""
        refseq_gi = self.esearch_request_one("nuccore", f"{refseq_id}[Accession]")
        if not refseq_gi:
            self.logger.warning(f"No RefSeq GI number for accession {refseq_id}")
            return None, None
        nuc_record = self.esummary_request_one("nuccore", refseq_gi)
        if not nuc_record:
            self.logger.warning(f"No nuccore record for RefSeq GI number {refseq_gi}")
            return None, None
        species = nuc_record.get("organism")
        taxid = nuc_record.get("taxid")
        if not taxid:
            self.logger.warning(f"No nuccore taxid for RefSeq GI number {refseq_gi}, trying assembly database")
            assembly_record = self.esummary_request_one("assembly", refseq_gi)
            species = assembly_record.get("organism", {}).get("scientificname", species)
            taxid = assembly_record.get("taxid")
        return species, taxid

    def download_zip_file(self, url, filename):
        """Download a zip file from the given URL and save it to the specified filename."""
        headers = {"accept": "application/zip"}

        def download_zip_file_core(url, filename):
            resp = requests.get(url, headers=headers, stream=True)
            if self.check_response(resp):
                with open(filename, "wb") as f:
                    for chunk in resp.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
                return filename
            else:
                raise Exception(f"Bad response from {url}")
        return self.query_with_retry(download_zip_file_core, url, filename)

    def post_download_zip_file(self, url, params, filename):
        """Download a zip file from the given URL and save it to the specified filename."""
        headers = {"accept": "application/zip"}

        def post_download_zip_file_core(url, params, filename):
            resp = requests.post(url, headers=headers, json=params, stream=True)
            if self.check_response(resp):
                with open(filename, "wb") as f:
                    for chunk in resp.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
                return filename
            else:
                raise Exception(f"Bad response from {url} with params:\n{json.dumps(params, indent=2)}")
        return self.query_with_retry(post_download_zip_file_core, url, params, filename)

    def download_refseq(self, assembly_id, filename):
        """Download RefSeq using the NCBI Datasets API which requires the GCF_* assembly ID."""
        url = f"{NCBI_DATASETS_V2_BASE}/genome/accession/{assembly_id}/download?include_annotation_type=GENOME_FASTA&include_annotation_type=GENOME_GBFF&filename={filename}"
        return self.download_zip_file(url, filename)

    def download_genbank(self, taxid, filename):
        """Download all GenBank genomes for the Taxonomy ID.  Return the .zip file name."""
        # Use the NCBI datasets command-line tool because the plain API seems to get cut off more often
        command = ["datasets", "download", "virus", "genome", "taxon", str(taxid),
                   "--filename", filename, "--no-progressbar"]
        return self.run_command(command)

    def download_genbank_accessions(self, accessions, filename):
        """Download specific GenBank accessions for the Taxonomy ID.  Return the .zip file name."""
        # If it looks weird that the URL has a GET-style param even though other params are separately POSTed,
        # I'm just following the API doc as of Aug. 2025:
        # https://www.ncbi.nlm.nih.gov/datasets/docs/v2/api/rest-api/#post-/virus/genome/download
        url = f"{NCBI_DATASETS_V2_BASE}/virus/genome/download?filename={filename}"
        params = {"include_sequence": ["GENOME"],
                  "aux_report": ["NONE"],
                  "accessions": accessions}
        return self.post_download_zip_file(url, params, filename)

    def query_ncbi_virus_metadata(self, taxid, filename):
        """Query the undocumented NCBI Virus API for metadata for all sequences for taxid.
        Unlike NCBI Datasets' data_report, this includes strain and serotype.  Results written to filename are CSV."""
        url = ("https://www.ncbi.nlm.nih.gov/genomes/VirusVariation/vvsearch2/" +
               "?fq=%7B%21tag%3DSeqType_s%7DSeqType_s%3A%28%22Nucleotide%22%29"
               f"&fq=VirusLineageId_ss%3A%28{taxid}%29&q=%2A%3A%2A&cmd=download&dlfmt=csv" +
               "&fl=accession%3AAccVer_s" +
               "%2Cisolate%3AIsolate_s" +
               "%2Ccountry_location%3ACountryFull_s" +
               "%2Cdate%3ACollectionDate_s" +
               "%2Clength%3ASLen_i" +
               "%2Cbiosample%3ABioSample_s" +
               "%2Chost%3AHost_s" +
               "%2Csubmitter%3ASubmitterAffilFull_s" +
               "%2Cauthors%3AAuthors_csv" +
               "%2Cstrain%3AStrain_s" +
               "%2Cserotype%3ASerotype_s" +
               "%2Csegment%3ASegment_s"
               )
        # Request the URL, write results to filename

        def query_ncbi_virus_metadata_core(url, filename):
            resp = requests.get(url, stream=True)
            if self.check_response(resp):
                with open(filename, "wb") as f:
                    for chunk in resp.iter_content(chunk_size=8192):
                        if chunk:
                            f.write(chunk)
                return filename
            else:
                raise Exception(f"Bad response from {url}")
        return self.query_with_retry(query_ncbi_virus_metadata_core, url, filename)

    def get_species_level_taxid(self, subspecies_taxid):
        """Return the species-level Taxonomy ID for a given subspecies Taxonomy ID."""
        # EFETCH with retmode=json returns only the subspecies ID, but I need the whole record with lineage etc.
        # That requires parsing the XML response.
        def get_species_level_taxid_core(subspecies_taxid):
            resp = requests.get(NCBI_EUTILS_EFETCH_URL, params={"id": subspecies_taxid, "db": "taxonomy"})
            if self.check_response(resp):
                # Parse the XML response to find the species-level Taxonomy ID
                root = ET.fromstring(resp.content)
                for taxon in root.findall(".//Taxon"):
                    if taxon.find("Rank").text == "species":
                        print(f"Found species-level Taxonomy ID: {taxon.find('TaxId').text} for subspecies Taxonomy ID: {subspecies_taxid}")
                        return taxon.find("TaxId").text
                return subspecies_taxid
            else:
                raise Exception(f"Bad response from {NCBI_EUTILS_EFETCH_URL} with params: id={subspecies_taxid}, db=taxonomy")
        return self.query_with_retry(get_species_level_taxid_core, subspecies_taxid)



import sys
import tomllib


DEFAULT_MIN_LENGTH_PROPORTION = 0.8
DEFAULT_MAX_N_PROPORTION = 0.25
DEFAULT_MAX_PARSIMONY = 1000
DEFAULT_MAX_BRANCH_LENGTH = 10000
DEFAULT_DOCKER_IMAGE = "angiehinrichs/viral_usher"


def read_config(config_path):
    try:
        with open(config_path, 'rb') as f:
            config = tomllib.load(f)
    except FileNotFoundError:
        print(f"Config file '{config_path}' does not exist.", file=sys.stderr)
        raise
    return config


def check_refseq_acc(refseq_acc):
    if not refseq_acc or not isinstance(refseq_acc, str):
        raise ValueError("RefSeq accession must be a non-empty string.")
    if not (refseq_acc.startswith("NC_") or refseq_acc.startswith("NZ_") or refseq_acc.startswith("AC_")):
        raise ValueError("RefSeq accession must start with 'NC_', 'NZ_' or 'AC_'.")


def check_taxonomy_id(taxonomy_id):
    if not taxonomy_id or not isinstance(taxonomy_id, str):
        raise ValueError("Taxonomy ID must be a non-empty string.")
    if not taxonomy_id.isdigit():
        raise ValueError("Taxonomy ID must be a numeric string.")


def check_refseq_assembly(refseq_assembly):
    if not refseq_assembly or not isinstance(refseq_assembly, str):
        raise ValueError("RefSeq assembly must be a non-empty string.")
    if not refseq_assembly.startswith("GCA_") and not refseq_assembly.startswith("GCF_"):
        raise ValueError("RefSeq assembly must start with 'GCA_' or 'GCF_'.")


def parse_config(config_path):
    """Read the config file and validate its contents."""
    config = read_config(config_path)
    try:
        check_refseq_acc(config['refseq_acc'])
        check_taxonomy_id(config['taxonomy_id'])
        check_refseq_assembly(config['refseq_assembly'])
    except ValueError as e:
        print(f"Error in config file {config_path}: {e}", file=sys.stderr)
        raise
    return config


def write_config(config, config_path):
    """Write a config TOML file to the specified path."""
    with open(config_path, 'w') as f:
        print(f"# viral_usher config for RefSeq {config['refseq_acc']}, taxonomy ID {config['taxonomy_id']}\n", file=f)
        for key in config:
            print(f"{key} = '{config[key]}'", file=f)
